#!/bin/bash

#####################################################
# This script will run vacuum analyze or 
# just analyze for any given table or list of tables.
# Only one iteration of the script will run at any
# given time.
# The script options are as follows:
#   -v vacuum
#   -a vacuum analyze
#   -vl run vacuum analyze and log
#   -al run analyze and log
# If only -l is passed by default the script will
# run an ANALYZE VERBOSE
# A sample call would be:
#   db_maint.sh -v operation_progress
#   db_maint.sh -a operation_progress backups
# Return codes for this scripts are as follows:
#   1 - bad parameters/options passed
#   2 - unable to access postgres
#   3 - error found
#   4 - script is already running
#   5 - psql command not found
#####################################################

SECONDS=0
script_name=${0##*/}

BPDIR=`/usr/bp/bin/bputil -g "Location Information" "BPDIR" "" /etc/default/bp.ini`
present_dir=`pwd`
out_dir="/var/log"
LOG="/dev/null"
maintenance_work_mem=200000
work_mem=100000
rc=0
run_option=""
run_verbiage="ANALYZE VERBOSE"
database_name="bpdb"
database_user="postgres"
table_list=""
max_workers_process_count=2
workers_process_count=0
timeout_seconds=1800
own_pid=$$
clk_tck=$(getconf CLK_TCK)

#
# Depending on the postgres version the column name to
# cancel a thread is different
#
cent_ver=`/bin/uname -r |sed -e 's/.*el//' |cut -c1`

#
#Try to find where psql is. If not found then exit
#
check_psql (){
/usr/bin/which psql > /dev/null 2>&1
rc=$?
if [[ $rc -eq 0 ]]; then
        psql_path="`/usr/bin/which psql`"
        psql_path=`echo ${psql_path%psql}`
        PSQL=`echo $psql_path/psql -v ON_ERROR_STOP=1`
else
        printf "Could not find psql command in the path, exiting now.\n" >> $LOG
	rc=5
	clean_up
fi
}

clean_up (){
        duration=$SECONDS
        duration=`echo $(($duration / 60))`
        printf "Script ran in ${duration} minutes...\n">>$LOG
	printf "RC=$rc\n" >> $LOG
	exit $rc
}

# Function to check if postgres is up
check_db_is_up (){
	rc=0
        sudo -iu postgres pg_ctl status -D /usr/bp/data >> $LOG 2>>$LOG
        rc=$?
	# This is necessary to use the function in the if statement below and get the right output
        echo $rc
}

check_rc (){
        if (( $rc != 0 )); then
                printf "Command returned an error...\nPlease check the $LOG for more info...\n">>$LOG
		printf "Command's returned code was $rc \n">>$LOG
                printf "Exiting now...\n">>$LOG
		rc=3
		clean_up
        fi
}

check_worker_process (){
	pid=$1
	uptime=`awk '{print $1}' < /proc/uptime`
	uptime=${uptime%.*}
	if [[ -e /proc/$pid/stat ]]; then
		starttime=`awk '{print $22}' /proc/$pid/stat 2>/dev/null`
		if [[ $starttime =~ '^[0-9]+$' ]]; then	
			(( starttime=starttime/clk_tck ))
			starttime=${starttime%.*}
			(( run_time=uptime-starttime ))
			if (( $run_time > $timeout_seconds )) && [[ -e /proc/$pid/stat ]]; then
				kill -15 $pid >> $LOG 2>&1
				printf "Had to kill process $pid because it exceeded $timeout_seconds seconds of runtime...\n">>$LOG
			fi
		fi
	fi
}

run_vacuum_analyze (){
	printf "BEGIN: `date +%F%H%M%S` \nRunning $run_verbiage for: ${table_list}\n" >> $LOG
	for table_name in $table_list
	do
		if [[ `check_db_is_up` -ne 0 ]]; then
			printf "Postgres seems to be down...\nExiting now...\n" >> $LOG
			printf "Please look at the $LOG log file for more details...\n" >> $LOG
			rc=2
			clean_up
		else
			if [ $cent_ver -ge 6 ]; then
				( /usr/bin/timeout -sSIGTERM ${timeout_seconds}s ${PSQL} -U $database_user $database_name <<-EOF >> $LOG 2>>$LOG
				SET MAINTENANCE_WORK_MEM TO ${maintenance_work_mem} ;
				$run_verbiage ${table_name} ;
				EOF
				rc=$?
				check_rc ) &
			else
				#
				# For cent5 send the maint script to background and spawn a second monitoring script
				# this is similar to how timeout actually works
				#
				(${PSQL} -U $database_user $database_name <<-EOF >> $LOG 2>>$LOG
				SET MAINTENANCE_WORK_MEM TO ${maintenance_work_mem} ;
				$run_verbiage ${table_name} ;
				EOF
				check_rc
				) &
			fi
		fi
		#workers_process_count=`ps -aef|grep -v grep|grep -v $own_pid|grep -c ${script_name}`
		workers_process_count=`/usr/bin/pgrep -P ${own_pid} |wc -l`
		while (( $workers_process_count > $max_workers_process_count ))
		do
			for pid in `/usr/bin/pgrep -P $own_pid`
			do
				check_worker_process $pid
			done
			usleep 10000
			workers_process_count=`/usr/bin/pgrep -P ${own_pid} |wc -l`
		done
	done

	workers_process_count=`/usr/bin/pgrep -P ${own_pid} |wc -l`
	while (( $workers_process_count > 1 ))
	do
		for pid in `/usr/bin/pgrep -P $own_pid`
		do
			check_worker_process $pid
		done
		sleep 5
		workers_process_count=`/usr/bin/pgrep -P ${own_pid} |wc -l`
	done
	wait
	printf "Done with all tables: ${table_list} \nEND: `date +%F%H%M%S` \n" >> $LOG
}

# main script
USAGE="Usage: $script_name [-fval] <table_list> \n -f Run VACUUM FULL ANALYZE VERBOSE\n -v Run VACUUM ANALYZE VERBOSE\n -a Run ANALYZE VERBOSE\n -l Log to /usr/bp/logs.dir\n"
lockFile=/var/lock/"${script_name}"-lock

# make sure the log file is present
touch $LOG
#
# If the lock file is older 24 hours give a Warning and exit
# Maybe a catastrophic event happend
#
if [[ -e ${lockFile} ]]; then
        if test "`find ${lockFile} -mmin +1440`"; then
                printf "Found ${lockFile}.  There is possibly a maintenance process still running...\n">>$LOG
                /usr/sbin/lsof -a ${lockFile}>>$LOG
		rc=4
                clean_up
        fi
fi

#
#  Make sure only one copy of this script can run at a time.
#
if ! lockfile -r0 "$lockFile"; then
        printf "${script_name}: Can't get the lock\n">>$LOG
	rc=4
	clean_up
fi

trap "rm -f $lockFile >/dev/null" EXIT

check_psql

if (( $# < 1 )); then
    printf "Invalid option passed.\n $USAGE"
        exit 1
fi

if [[ $1 == *"a"*"v"*\
   || $1 == *"a"*"f"*\
   || $1 == *"a"*"v"*"f"*\
   || $1 == *"a"*"f"*"v"*\
   || $1 == *"v"*"a"*\
   || $1 == *"v"*"f"*\ 
   || $1 == *"v"*"a"*"f"*\ 
   || $1 == *"v"*"f"*"a"*\ 
   || $1 == *"f"*"v"*\
   || $1 == *"f"*"a"*\ 
   || $1 == *"f"*"a"*"v"*\
   || $1 == *"f"*"v"*"a"* ]]; then
	printf "Invalid option passed.\n $USAGE"
	exit 1
fi

while getopts lvaf opt
do
	case "$opt" in
		v) run_option="v";run_verbiage="VACUUM ANALYZE VERBOSE";;
		f) run_option="f";run_verbiage="VACUUM FULL ANALYZE VERBOSE";;
		a) run_option="a";run_verbiage="ANALYZE VERBOSE";;
		l) LOG="${out_dir}/${script_name%.*}.log";
		   touch $LOG;;
		*) printf "Invalid option passed.\n $USAGE";
		   exit 1;;
	esac
done
shift $((OPTIND-1))

table_list="$*"
run_vacuum_analyze
