#!/bin/bash

#####################################################
# This script will run vacuum analyze 
# Only one iteration of the script will run at any
# given time.
#
# This script is schedule through 
# the /etc/cron.d/unitrends-pg file
#
# The script options are as follows:
#   -l run vacuum analyze and log
#   -m monitoring/mock run
#   -lm do a dry-run of the maintenance
#   -i ignore master.ini schedule and run
#   -f stop all bp service and start them after maint is done
# A sample call would be:
#   db_maint_driver.sh -l
#   db_maint_driver.sh -m
#   db_maint_driver.sh -lf
#   db_maint_driver.sh -lm
#   db_maint_driver.sh -lif
# Return codes for this scripts are as follows:
#   1 - bad parameters/options passed
#   2 - unable to access postgres
#   3 - error running psql command
#   4 - error found in log file
#   5 - scheduling mismatch
#   6 - psql command not found
#   7 - <currently not used>
#   8 - lock file found
#   9 - unable to find BPDIR directory or bputil executable
#   10 - process terminated by signal
#####################################################
# A good query to run against the history table is:
#  select to_timestamp(last_maint_time) as last_maint_time,to_timestamp(start_time) as start_time,
#  to_timestamp(end_time) as end_time,
#  last_elapsed_time,relname,reltuples,expectedpages,bloatpct,autovac_freeze_pct,current_status
#  from bp.bpdb_maint_driver_hst where relname = 'backups_hash_values' order by 1 desc;
#####################################################
# Per UNIBP-12311
# maximum number of cocurrent maintenance processes default is 3, min value is 1
#####################################################

export PATH=/sbin:/usr/sbin:/bin:/usr/bin:/usr/local/bin:/usr/bp/bin:/usr/bp/lib/shlib:/root/bin:/usr/local/pgsql/bin

script_name=${0##*/}
USAGE="Usage: $script_name [-lmif] \n \
	-l Run and log to /usr/bp/logs.dir\n \
	-m Run in monitor/mock mode\n \
	-lm Run in monitor mode and log\n \
	-i Run interactive ignore schedule\n \
	-f Stop/Start bp processes\n"
lockFile=/var/lock/"${script_name}"-lock
#
# default schedule settings if missing from the master.ini
#
default_scheduled_day="3"
default_scheduled_hour="11"
default_max_run_time="59"
default_max_maint_process_count="3"
default_last_maint_hr="24"

#
# Default sample and growth rate for performance analyzsis
#
default_sample_size=4
default_growth_rate=.5

#
# Max failed attempts before overwriting max_run_time
#
default_max_failed_attempts=3

#
# Ovewrite run time after max_failed_attempts is met
# Is in minutes 
# This is mainly used for default maintenance schedule
#
default_overwrite_run_time=300

# in interactive mode run upto 360 minutes or 6 hours
default_max_interactive_run_time=360

#
# minimum bloat percentage to run the maintenance
#
default_bloat_pct_threshold=20
#
# minimum autovac freeze xid consumed percentage to run the maintenance
#
default_xid_threshold=85
#
# minimum size in bytes ~ 100MB
#
default_rel_size_threshold=200000000

#
# number of days to keep records in the maint_hist table
#
default_hst_days_keep=365
#
# since the script can run from command line assume the script runs 24/7/365
# populating at this time (3/23/17) 232 rows per run
# if so, only keep 3 months worth of data
default_hst_max_rows=508080

#
# postgres session parameters
#
# maintenance_work_mem percentage of total memory available 
default_maintenance_work_mem=20
# work_mem in KB's
default_work_mem=512000
database_name="bpdb"
database_user="postgres"

#
# prefix of the tmp files and maintenance jobs
#
table_file_name_prefix="db_table_maint"

if [[ ! -x /usr/bp/bin/bputil ]]; then
	printf "Unable to find the bputil executable...\n Exiting now...\n"
	exit 9
fi
BPDIR=`/usr/bp/bin/bputil -g "Location Information" "BPDIR" "" /etc/default/bp.ini`
if [[  -z "$BPDIR" ]] || [[ ! -d "$BPDIR" ]]; then
	printf "Unable to find the BPDIR directory location...\n Exiting now...\n"
	exit 9
fi
database_dir=$BPDIR/data
present_dir=`pwd`
out_dir="${BPDIR}/logs.dir"
LOG="/dev/null"
rc=0
run_option=""
run_verbiage=""
current_epoch=`date '+%s'`
found_script=0
found_yum=0
timeout_seconds=0
failed_maintenance=0
table_size=0
last_elapsed_time=0
maintenance_space_needed=0
maintenance_space_check=0
stop_epoch=`date +%s`
current_day=`date '+%u'`
current_hour=`date '+%k'`
current_time=`date '+%k%M'`
restart_bp_fl="f"
pending_maint_fl="t"
pending_maint_count=0
pg_repack_fl="f"
has_pk_fl="f"
xid_maint_fl="f"
elapsed_maint_fl="f"
check_last_run_fl="f"
schedule_fl="f"
default_force_db_maint="f"
maint_command="VACUUM"
maint_process_count=0
maint_pid=$$
table_file_list=""
child_pid=""
scheduled_day=`$BPDIR/bin/bputil -g "DataBase" DBMaintDayOfWeek "${default_scheduled_day}" $BPDIR/bpinit/master.ini`
scheduled_hour=`$BPDIR/bin/bputil -g "DataBase" DBMaintHour "${default_scheduled_hour}" $BPDIR/bpinit/master.ini`
scheduled_run_time=`$BPDIR/bin/bputil -g "DataBase" DBMaintMaxRuntime "${default_max_run_time}" $BPDIR/bpinit/master.ini`
overwrite_run_time=`$BPDIR/bin/bputil -g "DataBase" DBMaintOverwriteRuntime "${default_overwrite_run_time}" $BPDIR/bpinit/master.ini`
max_interactive_run_time=`$BPDIR/bin/bputil -g "DataBase" DBMaintMaxInterRuntime "${default_max_interactive_run_time}" $BPDIR/bpinit/master.ini`
max_maint_process_count=`$BPDIR/bin/bputil -g "DataBase" DBMaintMaxProcessCount "${default_max_maint_process_count}" $BPDIR/bpinit/master.ini`
max_failed_attempts=`$BPDIR/bin/bputil -g "DataBase" DBMaintMaxFailed "${default_max_failed_attempts}" $BPDIR/bpinit/master.ini`
force_db_maint=`$BPDIR/bin/bputil -g "DataBase" DBMaintForce "${default_force_db_maint}" $BPDIR/bpinit/master.ini`
bloat_pct_threshold=`$BPDIR/bin/bputil -g "DataBase" DBMaintBloat "${default_bloat_pct_threshold}" $BPDIR/bpinit/master.ini`
xid_threshold=`$BPDIR/bin/bputil -g "DataBase" DBMaintXid "${default_xid_threshold}" $BPDIR/bpinit/master.ini`
maintenance_work_mem=`$BPDIR/bin/bputil -g "DataBase" DBMaintMem "${default_maintenance_work_mem}" $BPDIR/bpinit/master.ini`
work_mem=`$BPDIR/bin/bputil -g "DataBase" DBMaintWork "${default_work_mem}" $BPDIR/bpinit/master.ini`
rel_size_threshold=`$BPDIR/bin/bputil -g "DataBase" DBMaintTblSize "${default_rel_size_threshold}" $BPDIR/bpinit/master.ini`
hst_days_keep=`$BPDIR/bin/bputil -g "DataBase" DBMaintDaysKeep "${default_hst_days_keep}" $BPDIR/bpinit/master.ini`
hst_max_rows=`$BPDIR/bin/bputil -g "DataBase" DBMaintRowsKeep "${default_hst_max_rows}" $BPDIR/bpinit/master.ini`
sample_size=`$BPDIR/bin/bputil -g "DataBase" DBMaintSampleSize "${default_sample_size}" $BPDIR/bpinit/master.ini`
growth_rate_threshold=`$BPDIR/bin/bputil -g "DataBase" DBMaintGrowthRate "${default_growth_rate}" $BPDIR/bpinit/master.ini`
last_maint_hr=`$BPDIR/bin/bputil -g "DataBase" DBMaintLast "${default_last_maint_hr}" $BPDIR/bpinit/master.ini`
# convert the run_times from master.ini to seconds
(( overwrite_run_time=overwrite_run_time*60 ))
(( scheduled_run_time=scheduled_run_time*60 ))
(( max_interactive_run_time=max_interactive_run_time*60 ))

# convert maintenance_work_mem into a percent
maintenance_work_mem=$(echo "scale=2; $maintenance_work_mem / 100" |bc)

#
# Function to put a standard exit message
# when a signal is detected and to remove the lockfile
#
cleanup_sig (){
	#
	# Restart bp services if necessary
	#
	if [[ "${restart_bp_fl}" == "t" ]]; then
		start_services
	fi

	printf "WARNING: Process caught a signal...\n" >> $LOG
	printf "INFO: Attempting to run pkill -TERM -P $$ \n" >> $LOG
	/usr/bin/pkill -TERM -P $$ >> $LOG 2>&1
	#
	# If we still have any VACUUM process running kill it
	#
	printf "INFO: Attempting to cancel any leftover VACUUM or REINDEX operations after signal caught\n" >> $LOG
	$PSQL -U $database_user $database_name -tqc"SELECT pg_cancel_backend($pid_column) FROM pg_stat_activity WHERE upper($current_query_column) like 'VACUUM%'
	OR upper($current_query_column) like 'REINDEX%' " >> $LOG 2>&1
	if [ $cent_ver -ge 6 ]; then
		$PSQL -U $database_user $database_name -tqc"SELECT pg_terminate_backend($pid_column) FROM pg_stat_activity WHERE 
		upper($current_query_column) like 'VACUUM%' OR upper($current_query_column) LIKE 'REINDEX%' " >> $LOG 2>&1
	fi
	/bin/rm -f $lockFile /tmp/${table_file_name_prefix}* >/dev/null 2>&1
	printf "Exiting now...\n" >> $LOG
	printf "END: `date` \n" >> $LOG 
	exit 10
}

trap cleanup_sig SIGHUP SIGINT SIGTERM

#
# Depending on the postgres version the column name to
# cancel a thread is different
#
cent_ver=`/bin/uname -r |sed -e 's/.*el//' |cut -c1`

if [ $cent_ver -ge 6 ]; then
        pid_column="pid"
	current_query_column="query"
else
        pid_column="procpid"
	current_query_column="current_query"
fi

#
# Function to put a standard exit message
# when exiting and to remove the lockfile
#
cleanup (){
	# Restart bp services if necessary
	if [[ "${restart_bp_fl}" == "t" ]]; then
		start_services
	fi

	printf "Exiting now...\n" >> $LOG
	printf "END: `date` \n" >> $LOG 
	/bin/rm -f $lockFile /tmp/${table_file_name_prefix}* >/dev/null 2>&1
}

#
#Try to find where psql is. If not found then exit
#
check_psql (){
/usr/bin/which psql > /dev/null 2>&1
rc=$?
if [[ $rc -eq 0 ]]; then
	psql_path="`/usr/bin/which psql`"
	psql_path=`echo ${psql_path%psql}`
	PSQL=`echo $psql_path/psql`
else
	printf "Could not find psql command in the path, exiting now.\n" >> $LOG
	cleanup
	printf "Exit 6\n" >> $LOG
	exit 6
fi
}

#
# Function used to check that the return code from a psql
# statement was successful otherwise exit
#
check_sql_rc (){
	if [[ $rc -ne 0 ]]; then
		printf "Error: Unable to execute sql in the $funcname \n" >> $LOG
		cleanup
		printf "Exit 3\n" >> $LOG
		exit 3
	fi
}

#
# This function create a populate_driver_tbl postgres function
# Is used to refresh stats information for all bpdb tables
#
populate_driver_table (){
	funcname=$FUNCNAME
	if [ $cent_ver -ge 6 ]; then
		${PSQL} -U $database_user $database_name <<-EOF >>$LOG 2>&1
		SET search_path TO bp;
		SELECT bp.populate_driver_tbl();
		ANALYZE VERBOSE bp.bpdb_maint_driver;
		\q
		EOF
		rc=$?
		check_sql_rc
	else
		${PSQL} -U $database_user $database_name <<-EOF >>$LOG 2>&1
		SET search_path TO bp;
		SELECT bp.populate_driver_tbl_pg8();
		ANALYZE VERBOSE bp.bpdb_maint_driver;
		\q
		EOF
		rc=$?
		check_sql_rc
	fi
	#
	# Use the estimated partition throughput in KB/s and figure out
	# more or less how long it could take to run maintenance per table
	# adding a cushion of 30x to account for other workloads
	#
	$PSQL -U $database_user $database_name <<-EOF >> $LOG 2>&1
	SET SEARCH_PATH to bp;
	UPDATE bp.bpdb_maint_driver SET \
	estimated_elapsed_time=((relpages*8192)/${estimated_db_performance})*30;
	\q
	EOF
	#
	# Add all estimated elapsed times and if they are more than the current
	# DBMaintMaxInterRuntime setting in the master.ini then update the setting
	# And recalculate the stop_epoch
	sum_estimated_elapsed_time=`$PSQL -U $database_user $database_name -tc"SELECT sum(estimated_elapsed_time) FROM bp.bpdb_maint_driver"`
	sum_estimated_elapsed_time=${sum_estimated_elapsed_time# }
	sum_estimated_elapsed_time=${sum_estimated_elapsed_time% }
	if (( ${sum_estimated_elapsed_time} > ${max_interactive_run_time} )); then
		# 2x the DBMaintMaxInterRuntime and make sure to convert it back to minutes and then seconds again
		(( max_interactive_run_time=2*sum_estimated_elapsed_time ))
		(( max_interactive_run_time=max_interactive_run_time/60 ))
		$BPDIR/bin/bputil -p "DataBase" DBMaintMaxInterRuntime "${max_interactive_run_time}" $BPDIR/bpinit/master.ini
		(( max_interactive_run_time=max_interactive_run_time*60 ))
		# Recalculate the stop and time left variables
		(( stop_epoch=stop_epoch+max_interactive_run_time ))
		stop_epoch_date=`date -d @${stop_epoch}`
		printf "WARNING: Recalculated maintenance stop time based on estimated elapsed time. Maintenance will now stop by $stop_epoch_date \n" >> $LOG
	fi
}

#
# This function takes all information from the bpdb_maint_driver table
# and after inserting it into the hst table it truncates it
#
populate_hst_table (){
	funcname=$FUNCNAME
	${PSQL} -U $database_user $database_name <<-EOF >>$LOG 2>&1
	SET search_path TO bp;
	ANALYZE VERBOSE bp.bpdb_maint_driver_hst;
	SELECT bp.cleanup_maint_hst_tbl(${hst_days_keep},${hst_max_rows});
	SELECT bp.populate_hst_tbl();
	ANALYZE VERBOSE bpdb_maint_driver;
	-- Run a vacuum full analyze after purge
	VACUUM ANALYZE VERBOSE bpdb_maint_driver_hst;
	\q
	EOF
	rc=$?
	check_sql_rc
}

#
# This function checks the master.ini file values for the db maintenance
# Make sure they are valid and it makes sure it really is time to run or not
#
check_schedule (){
	funcname=$FUNCNAME
	if [ "$scheduled_day" -ge 0 -a "$scheduled_day" -le 6 ]; then
		if (( $scheduled_day != $current_day )); then
			printf "Scheduled maintenance day is ${scheduled_day}. Current day is ${current_day}.\n" >> $LOG
			schedule_fl="f"
		fi
	else
		printf "Invalid value for DBMaintDayOfWeek= ${scheduled_day}. Valid values are between 0-6.\n" >> $LOG
		schedule_fl="f"
	fi
	if [ "$scheduled_hour" -ge 0 -a "$scheduled_day" -le 23 ]; then
		if (( $scheduled_hour != $current_hour )); then
			printf "Scheduled maintenance hour is ${scheduled_hour}. Current hour is ${current_hour}.\n" >> $LOG
			schedule_fl="f"
		fi
	else
		printf "Invalid value for DBMaintHour is ${scheduled_hour}. Valid values are between 0-23.\n" >> $LOG
		schedule_fl="f"
	fi
	if [ "$scheduled_run_time" -ge 10 -a "$scheduled_run_time" -le $max_interactive_run_time ]; then
		if (( $current_epoch > $stop_epoch )); then
			printf "Scheduled stop time is $scheduled_time. Current time is ${current_time}.\n" >> $LOG
			schedule_fl="f"
		fi
	else
		printf "Invalid value for DBMaintMaxRuntime is ${scheduled_run_time}. Valid values are between 10-${max_interactive_run_time} minutes\n" >> $LOG
		schedule_fl="f"
	fi
	# Don't allow more than 6 concurrent vacuum processes to run
	if [ "$max_maint_process_count" -lt 1 -o "$max_maint_process_count" -gt 6 ]; then
		printf "Current max process count is ${max_maint_process_count}.\n" >> $LOG
		printf "Invalid value for DBMaintMaxProcessCount. Valid values are between 1-6.\n" >> $LOG
		schedule_fl="f"
	fi
}

kill_maint (){
	funcname=$FUNCNAME
	printf "WARNING: Killing any left over maintenance processes now...\n" >> $LOG
	for pid_number in `/bin/ps -aef|grep -v grep|grep "${table_file_name_prefix}"|awk '{print $2}'`
	do
		temp_name=`/bin/ps -fp $pid_number|tail -1`
		temp_name=${temp_name##*${table_file_name_prefix}_}
		schema_name=${temp_name%%.*}
		table_name=${temp_name##*.}
		table_name=${table_name%_*}
		child_pid=`/usr/bin/pgrep -P ${pid_number} |tail -1`
		# Go through any child process foudn and try to kill them one at a time
		printf "WARNING: Found $pid_number Trying to kill its child processes...\n" >> $LOG
		/bin/ps -ef|grep $pid_number >> $LOG
		while [[ $child_pid != "" ]]; do
			printf "INFO: Killing $child_pid of $pid_number ...\n" >> $LOG
			/bin/kill -TERM $child_pid >> $LOG 2>&1
			child_pid=`/usr/bin/pgrep -P ${pid_number} |tail -1`
		done
		printf "INFO: Running kill ${pid_number} ...\n" >> $LOG	
		/bin/kill $pid_number >>$LOG 2>&1
		#
		# Try to update the status in the bpdb_maint_driver table
		#
		$PSQL -U $database_user $database_name -tqc"UPDATE bp.bpdb_maint_driver SET current_status = 't', end_time = extract(epoch from now())::bigint,
			comment='-Maintenance exceeded runtime-'||comment
			WHERE relname = '${table_name}' and schemaname='${schema_name}';" >>$LOG 2>&1
		check_sql_rc
		sleep 5
	done
	#
	# If we still have any VACUUM process running kill it
	#
	printf "INFO: Attempting to cancel any leftover VACUUM or REINDEX operations after maintenance time exceeded\n" >> $LOG
	$PSQL -U $database_user $database_name -tqc"SELECT pg_cancel_backend($pid_column) FROM pg_stat_activity WHERE upper($current_query_column) like 'VACUUM%'
	OR upper($current_query_column) LIKE 'REINDEX%' " >> $LOG 2>&1
	if [ $cent_ver -ge 6 ]; then
		$PSQL -U $database_user $database_name -tqc"SELECT pg_terminate_backend($pid_column) FROM pg_stat_activity WHERE 
		upper($current_query_column) like 'VACUUM%' OR upper($current_query_column) LIKE 'REINDEX%' " >> $LOG 2>&1
	fi

}
#
# This function makes sure no ongoing maintenance runs past
# the max_run_time. Otherwise it will kill the process and 
# update the status of the maint for the table in question in the
# bpdb_maint_driver table
#
check_maint_processes (){
	funcname=$FUNCNAME
	current_epoch=`date '+%s'`
	if (( $current_epoch > $stop_epoch )); then
		kill_maint
	fi
}

check_bpdb_is_available (){
        wait_time=0
        ready_found=0
        recovery_found=0
        recovery_msg=0
        total_wait_time=180
        postgres_log_ln=0
        current_postgres_log_ln=0
        postgres_log=`ls -t $database_dir/pg_log|awk '{print $1}'|head -1`
        database_name="bpdb"
        # Wait no more than 5 minutes or 300 seconds for the datbase to be ready
        ${PSQL} -U $database_user $database_name -c"select 1 from bp.backups limit 1" > /dev/null 2>&1 
        ready_found=$?
        if [[ $ready_found != 0 ]]; then
                current_postgres_log_ln=`wc -l $BPDIR/data/pg_log/$postgres_log|awk '{print $1}'`
                while (( $wait_time < $total_wait_time ))
                do
                        postgres_log_ln=`wc -l $database_dir/pg_log/$postgres_log|awk '{print $1}'`
                        (( postgres_log_ln_diff=current_postgres_log_ln-postgres_log_ln ))
                        # Look for certain keywords if found exit
                        tail_lines=`echo "-"$postgres_log_ln_diff`
                        printf "Running select statement to check database availability...\n" >> $LOG
                        ${PSQL} -U $database_user $database_name -c"select 1 from bp.backups limit 1" > /dev/null 2>&1
                        ready_found=$?
                        # if we haven't found the "is ready" message keep waiting
                        if [[ $ready_found == 0 ]]; then
                                printf "Database is now available...\n" >> $LOG
                                break;
                        fi
                        # Look for recovery mode and just display a message once
                        tail $tail_lines $database_dir/pg_log/$postgres_log | grep -q "recovery in progress"
                        recovery_found=$?
                        if [[ $recovery_found == 0 ]] && (( $recovery_msg == 0 )); then
                                echo "NOTICE: Database is starting up.  Please stand-by." >> $LOG
                                recovery_msg=1
                                total_wait_time=300
                        fi
                        sleep 20
                        (( wait_time=wait_time+20 ))
                done
        fi
        # if no message is found, postgres could be in recovery pending for a long time and we need to stop
        ${PSQL} -U $database_user $database_name -c"select 1 from bp.backups limit 1" > /dev/null 2>&1 
        ready_found=$?
		if [[ $ready_found != 0 ]]; then
			echo "ERROR: Please verify the $database_dir/pg_log/$postgres_log log. Database seems to be down, aborting..." >> $LOG
			# Restart bp services if FORCE was not used since cleanup will not do it
			if [[ "${restart_bp_fl}" == "t" ]]; then
				printf "WARNING: Will attempt to start all bp services now...\n" >> $LOG
				# Attempt to restart all services because DBMaintForce was specified
				# We would have stop_services prior
				start_services
			fi
			cleanup
			exit 2
		fi
}

# Stop all bp services
stop_services (){
        printf "Running /usr/bp/bin/stop_db.sh...\n" >> $LOG
        /usr/bp/bin/stop_db.sh >> $LOG 2>>$LOG

        printf "Stopping all processes...\n" >> $LOG
        /etc/init.d/bp_rcscript stop >> $LOG 2>>$LOG

        printf "Stopping httpd processes...\n" >> $LOG
        service httpd stop >> $LOG 2>>$LOG

        printf "Stopping crond processes...\n" >> $LOG
        service crond stop >> $LOG 2>>$LOG
}

# Start all bp services
start_services (){
        printf "Starting httpd processes...\n" >> $LOG
        service httpd start >> $LOG 2>>$LOG

        printf "Starting crond processes...\n" >> $LOG
        service crond start >> $LOG 2>>$LOG

        printf "Starting all processes...\n" >> $LOG
        /etc/init.d/bp_rcscript start >> $LOG 2>>$LOG
}

#
# The xid_maint and light_maint functions are used to
# build the necessasry maint commands on the tmp maint files
#
check_maint_rc (){
	printf "if (( \$rc != 0 )); then \n" >> $table_file_name
	printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET current_status = 'f', end_time = extract(epoch from now())::bigint , \
		comment = '-%s command failed-'||comment \
		WHERE schemaname = '%s' AND relname = '%s';\" \n" \
		"$PSQL" "$database_user" "$database_name" "${maint_command}" "${schemaname}" "${table_name_only}" >> $table_file_name
	printf "exit 0 \n" >> $table_file_name
	printf "else \n" >> $table_file_name
	printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET current_status = 's', end_time = extract(epoch from now())::bigint, \
		comment = '-%s command was successful-'||comment \
		WHERE schemaname = '%s' AND relname = '%s';\" \n" \
		"$PSQL" "$database_user" "$database_name" "${maint_command}" "${schemaname}" "${table_name_only}" >> $table_file_name
	printf "fi \n" >> $table_file_name
}
vacuum_maint (){
	# The start_time epoch was set before building this statement
	maint_command="VACUUM ANALYZE VERBOSE"
	printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET \
		start_time = ${start_time},comment = '-Running %s command-'||comment \
		WHERE schemaname = '%s' AND relname = '%s'\" \n" \
		"$PSQL" "$database_user" "$database_name" "${maint_command}" "${schemaname}" "${table_name_only}" >> $table_file_name
	printf "%s -U %s %s -tqac\"%s %s\" >> $LOG 2>&1 \n" \
		"$PSQL" "$database_user" "$database_name" "${maint_command}" "$table_name">> $table_file_name
	printf "rc=\$\? \n" >> $table_file_name
	check_maint_rc
}

reindex_maint (){
	# The start_time epoch was set before building this statement
	maint_command="REINDEX TABLE"
	printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET \
		start_time = ${start_time},comment = '-Running ${maint_command} command-'||comment \
		WHERE schemaname = '%s' AND relname = '%s';\" \n" \
		"$PSQL" "$database_user" "$database_name" "${schemaname}" "${table_name_only}" >> $table_file_name
	printf "%s -U %s %s -tqac\"%s %s \" >> $LOG 2>&1 \n" "$PSQL" "$database_user" "$database_name" "$maint_command" "$table_name">> $table_file_name
	printf "rc=\$\? \n" >> $table_file_name
	check_maint_rc
}

vacuum_full_maint (){
	# The start_time epoch was set before building this statement
	maint_command="VACUUM FULL ANALYZE VERBOSE"
	printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET \
		start_time = ${start_time},comment = '-Running %s command-'||comment \
		WHERE schemaname = '%s' AND relname = '%s'\" \n" \
		"$PSQL" "$database_user" "$database_name" "${maint_command}" "${schemaname}" "${table_name_only}" >> $table_file_name
	printf "%s -U %s %s -tqa <<-EOF >> $LOG 2>&1 \n" "$PSQL" "$database_user" "$database_name" >> $table_file_name
	printf "SET MAINTENANCE_WORK_MEM TO %s;\n" "$maintenance_work_mem" >> $table_file_name
	printf "SET WORK_MEM TO %s;\n" "$work_mem" >> $table_file_name
	printf "%s %s ;\n" "${maint_command}" "$table_name" >> $table_file_name
	printf "\\q \n" >> $table_file_name # Use escpae the backslash
	printf "EOF\n" >> $table_file_name #No spaces between EOF and new-line
	printf "rc=\$\? \n" >> $table_file_name
	check_maint_rc
}

light_maint (){
	reindex_maint
	vacuum_maint
}

check_last_run (){
	check_last_run_fl=`$PSQL -U $database_user $database_name -tqc \
        "SELECT EXISTS(SELECT 1 FROM bp.bpdb_maint_driver_hst WHERE 
	schemaname='${schemaname}' AND relname = '${table_name_only}' AND current_status = 's')"|head -1`
	check_last_run_fl=${check_last_run_fl# }

	if [[ "${check_last_run_fl}" == "t" ]]; then
		#
		# Check to see if last maint process ran more than xx hours ago
		#
		check_last_run_fl=`$PSQL -U $database_user $database_name -tqc \
		"WITH last_successful_run AS
		(SELECT MAX(end_time) as last_run FROM bp.bpdb_maint_driver_hst WHERE 
		schemaname='${schemaname}' AND relname = '${table_name_only}' AND current_status = 's')
		SELECT EXISTS(SELECT 1 FROM last_successful_run WHERE
		last_run < (date_part('epoch'::text,now() - interval '${last_maint_hr} hour')))"|head -1`
		check_last_run_fl=${check_last_run_fl# }
	else
		#
		# set check_last_run_fl to t because we don't have any prior succesful runs
		#
		check_last_run_fl="t"
	fi
}

check_repack (){
	#
	# Check to see if pg_repack is installed
	# If it is not then run VACUUM
	# Don't run pg_repack if we are stopping bp processes
	#
	pg_repack_fl=`$PSQL -U $database_user $database_name -tqc \
	"SELECT EXISTS(SELECT 1 FROM pg_extension WHERE extname = 'pg_repack')"|head -1`
	pg_repack_fl=${pg_repack_fl# }
	if [[ "${pg_repack_fl}" == "t" && "${restart_bp_fl}" != "t" ]]; then
		#
		# Check if table has primary key to use pg_repack
		# If it doesn't we need to do a VACUUM FULL
		#
		has_pk_fl=`$PSQL -U $database_user $database_name -tqc \
		"SELECT c.relhaspkey FROM pg_catalog.pg_class c JOIN pg_namespace n ON (c.relnamespace = n.oid
		AND n.nspname = '${schemaname}' AND c.relkind='r') AND relname='${table_name_only}'
		ORDER BY c.relhaspkey, c.relname"|head -1`
		has_pk_fl=${has_pk_fl# }
		if [[ "${has_pk_fl}" == "t" ]]; then
			printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET \
			comment = '-Running /usr/local/pgsql/bin/pg_repack -j 2 -n -D --table command-'||comment \
			WHERE schemaname = '%s' AND relname = '%s';\" \n" \
			"$PSQL" "$database_user" "$database_name" "${schemaname}" "${table_name_only}" >> $table_file_name
			printf "/usr/local/pgsql/bin/pg_repack -j 2 -n -D -U %s -d %s --table %s >> $LOG 2>&1 \n" \
			"$database_user" "$database_name" "$table_name" >> $table_file_name
		else
			# Set the start_time in epoch format now
			# in case we run multiple update commands
			start_time=`date '+%s'`
			if [[ "${xid_maint_fl}" == "t" ]]; then
				vacuum_maint
			else
				vacuum_full_maint
			fi
		fi
	else
		# Set the start_time in epoch format now
		# in case we run multiple update commands
		start_time=`date '+%s'`
		if [[ "${xid_maint_fl}" == "t" ]]; then
			vacuum_maint
		else
			vacuum_full_maint
		fi
	fi
}

run_table_maint (){
	#
	# Create a bash script file in /tmp to run the maintenance on the selected table
	#
	schemaname=${table_name%.*}
	table_name_only=${table_name#*.}
	if [[ "${run_option}" =~ ^(i|m)$ ]]; then
		table_name_list="${table_name} ${table_name_list}"
		printf "The $table_name table qualifies for maintenance.\n"|tee -a $LOG
		$PSQL -U $database_user $database_name <<- EOF |tee -a $LOG
		\qecho Last Known Stats:
		SELECT relname,reltuples as num_rows,
		pg_size_pretty(total_rel_size) as rel_size,
		relpages,expectedpages,
		bloatpct,autovac_freeze_pct
		FROM bp.bpdb_maint_driver_hst
		WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}'
		AND start_time = (select start_time from bp.bpdb_maint_driver_hst
			WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}' order by start_time desc limit 1);
		EOF
	fi

	# 
	# Figure out how many seconds can the maintenance job actually run without going over the stop time
	#
	(( timeout_seconds=stop_epoch-current_epoch ))

	#
	# Run a query to gage the current performance of the table before maintenance
	# Since it can take a long time, run it in background subshell
	#
	(
	/usr/bin/timeout $timeout_seconds ${PSQL} -U ${database_user} ${database_name} \
	-c"select bp.table_performance('${schemaname}','${table_name_only}')" >> $LOG 2>&1
	) &

	#
	# If the run_option was m (maintenance) then don't execute the bash file
	#
	if [[ "${run_option}" != "m" ]]; then
		#
		# Call check_last_run to set the check_last_run_fl flag
		# if the value is f then don't run any maintenance
		#
		check_last_run
		if [[ "${check_last_run_fl}" != "t" ]]; then
			printf "INFO: Last successful maintenance for ${table_name} was not more than $last_maint_hr hours ago\n" >> $LOG
			printf "INFO: Maintenance will not run for ${table_name}\n" >> $LOG
			$PSQL -U $database_user $database_name <<-EOF >> $LOG 2>&1
			SET SEARCH_PATH to bp;
			UPDATE bp.bpdb_maint_driver SET current_status = 'n',
			end_time=(date_part('epoch'::text, now()))::bigint,
			comment='-Maintenance not needed, last maint was ${last_maint_hr} hours ago-' || comment
			WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';
			\q
			EOF
			#
			# Exit the function now
			#
			return 0
		fi

		table_file_name=/tmp/${table_file_name_prefix}_${table_name}_${RANDOM}

		#
		# Start creating the tmp script to run the maintenance on the table
		#
		/bin/touch $table_file_name
		printf "#!/bin/bash \n" > $table_file_name
		#
		# This is a clean_up function used executed if the timeout is met we will mark the maintenance as failed
		#
		printf "function clean_up {\n" >> $table_file_name
		printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET current_status = 't', end_time = extract(epoch from now())::bigint, \
			comment = '-Maintenance terminated by kill command-'|| comment \
			WHERE schemaname = '%s' AND relname = '%s';\" >> $LOG 2>&1 \n" \
			"$PSQL" "$database_user" "$database_name" "${schemaname}" "${table_name_only}" >> $table_file_name
		printf "echo INFO: Maintenance terminated by kill command for \$\$ >> $LOG 2>&1 \n" >> $table_file_name
		printf "/usr/bin/pkill -TERM -P \$\$ >> $LOG 2>&1 \n" >> $table_file_name
		printf " exit 1 \n" >> $table_file_name
		printf "}\n" >> $table_file_name

		#
		# trap TERM and execute the clean_up function
		#
		printf "trap clean_up SIGHUP SIGINT SIGTERM \n" >> $table_file_name

		printf "rc=0 \n" >> $table_file_name

		#
		# Check if we are running any special maint
		#
		if [[ "${xid_maint_fl}" == "t" ]]; then
			printf "INFO: Running VACUUM ANALYZE VERBOSE for ${table_name} because of xid...\n" >> $LOG
			#
			# Call check_repack in case pg_repack was installed
			#
			check_repack
		elif [[ "${elapsed_maint_fl}" == "t" ]]; then
			# Set the start_time in epoch format now
			# in case we run multiple update commands
			printf "INFO: Running REINDEX and VACUUM ANALYZE VERBOSE for ${table_name} because of performance...\n" >> $LOG
			start_time=`date '+%s'`
			light_maint
		else
			#
			# Check if we have enough space for either VACUUM or pg_repack
			# If we don't have enough make a note and move on
			db_repo_used_kb=$(df -Pk $db_repo | tail -1 | awk '{print $3}')
			db_repo_free_kb=$(df -Pk $db_repo | tail -1 | awk '{print $4}')
			table_size=`$PSQL -U $database_user $database_name -tqc \
			"SELECT total_rel_size FROM bp.bpdb_maint_driver WHERE schemaname = '${schemaname}' AND 
			relname = '${table_name_only}'"|head -1`
			table_size=${table_size# }
			#
			# convert table_size to KB
			#
			(( table_size=table_size/1024 ))
			#
			# We need 2x the space to run VACUUM FULL or pg_repack
			#
			(( maintenance_space_needed=table_size*2 ))
			(( maintenance_space_check=db_repo_free_kb+maintenance_space_needed ))

			#
			# If we don't have enough space then need to run a light maint of reindex and vacuum only
			#
			if (( $maintenance_space_check > $db_repo_total_kb )); then
				printf "WARNING: Not enough space in $db_repo to run VACUUM FULL maintenance on ${table_name}.\n" >> $LOG 
				printf "INFO: Will attempt to run REINDEX and VACUUM on ${table_name}.\n" >> $LOG 
				# Set the start_time in epoch format now
				# in case we run multiple update commands
				start_time=`date '+%s'`
				light_maint
			else
				#
				# Prior to pg9, it is worth it to REINDEX before running a VACUUM
				#
				if [ $cent_ver -lt 6 ]; then
					# Set the start_time in epoch format now
					# in case we run multiple update commands
					start_time=`date '+%s'`
					reindex_maint
					#
					# Don't want to set maintenance_work_mem or work_mem
					#
					maint_command="VACUUM FULL ANALYZE VERBOSE"
					printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET comment = '-Running %s command-'||comment \
						WHERE schemaname = '%s' AND relname = '%s'\" \n" \
						"$PSQL" "$database_user" "$database_name" "${maint_command}" "${schemaname}" "${table_name_only}" >> $table_file_name
					printf "%s -U %s %s -tqac\"%s %s\" >> $LOG 2>&1 \n" \
						"$PSQL" "$database_user" "$database_name" "${maint_command}" "$table_name">> $table_file_name
				else
					#
					# Call check_repack in case pg_repack was installed
					#
					check_repack
				fi
			fi
		fi
		/bin/chmod 755 $table_file_name
		current_epoch=`date '+%s'`

		#
		# Figure out how many seconds can the maintenance job actually run without going over the stop time
		#
		(( timeout_seconds=stop_epoch-current_epoch ))
		#
		# If the diff is negative then don't run, we went over
		# set the current status to deferred
		#
		if (( $timeout_seconds > 0 )); then
			# If the table failed to run maintenance based on DBMaintMaxFailed in master.ini consecutive times
			# Increase the timeout
			failed_maintenance=`$PSQL -U $database_user $database_name -c"SELECT bp.report_bpdb_maint('${schemaname}','${table_name_only}') \
						LIMIT ${max_failed_attempts};"|grep -wc 'f'`
			if (( $failed_maintenance >= $max_failed_attempts )); then
				# Check the last_elapsed_time and 3x the last_elapsed_time if necessary
				# Make sure we return a positive number if 0 or NULL, default to 300 seconds or 5 minutes
				last_elapsed_time=`$PSQL -U $database_user $database_name -tqc"SELECT CASE WHEN (last_elapsed_time=0 OR last_elapsed_time IS NULL) \
							THEN 300 ELSE last_elapsed_time END FROM bp.bpdb_maint_driver \
							WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';"`
				# Cleanup white spaces
				last_elapsed_time=${last_elapsed_time# }
				last_elapsed_time=${last_elapsed_time% }
				# Use the estimated_elapsed_time if applicable
				# Otherwise triple the last_elapsed_time and use it
				estimated_elapsed_time=`$PSQL -U $database_user $database_name -tqc"SELECT CASE WHEN (estimated_elapsed_time=0 OR estimated_elapsed_time IS NULL) \
							THEN 300 ELSE estimated_elapsed_time END FROM bp.bpdb_maint_driver \
							WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';"`
				estimated_elapsed_time=${estimated_elapsed_time# }
				estimated_elapsed_time=${estimated_elapsed_time% }
				if (( $last_elapsed_time < $estimated_elapsed_time )); then
					last_elapsed_time=$estimated_elapsed_time
				else
					(( last_elapsed_time=3*last_elapsed_time ))
				fi
				# if 3xlast_elapsed_time is larger than the DBMaintOverwriteRuntime in master.ini overwrite it
				if (( $last_elapsed_time > $overwrite_run_time )); then
					# The last_elapsed_time is will now be the new ovewrite_run_time
					$PSQL -U $database_user $database_name <<-EOF >> $LOG 2>&1
					SET SEARCH_PATH to bp;
					UPDATE bp.bpdb_maint_driver SET comment = '-Maintenance run time adjusted because of \
					${max_failed_attempts} previously failed runs. Changed max elapsed time from ${overwrite_run_time} seconds to \
					${last_elapsed_time} -'||comment \
					WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';
					\q
					EOF
					overwrite_run_time=$last_elapsed_time
					#
					# if necessary also update the DBMaintMaxInterRuntime in master.ini for the next run 2x
					# Make sure to convert the seconds to minutes before updating master.ini and back
					# to seconds to use in the script
					#
					if (( $max_interactive_run_time < $overwrite_run_time )); then
						(( max_interactive_run_time=2*overwrite_run_time ))
						(( max_interactive_run_time=max_interactive_run_time/60 ))
						$BPDIR/bin/bputil -p "DataBase" DBMaintMaxInterRuntime "${max_interactive_run_time}" $BPDIR/bpinit/master.ini
						(( max_interactive_run_time=max_interactive_run_time*60 ))
						# Recalculate the stop and time left variables
						(( stop_epoch=stop_epoch+max_interactive_run_time ))
						stop_epoch_date=`date -d @${stop_epoch}`
						(( timeout_seconds=stop_epoch-current_epoch ))
						printf "WARNING: Recalculated maintenance stop time based on last_elapsed_time. Maintenance will now stop by $stop_epoch_date \n" >> $LOG
					fi
					(( overwrite_run_time=overwrite_run_time/60 ))
					$BPDIR/bin/bputil -p "DataBase" DBMaintOverwriteRuntime "${overwrite_run_time}" $BPDIR/bpinit/master.ini
					(( overwrite_run_time=overwrite_run_time*60 ))
				fi
				printf "NOTICE: The last $max_failed_attempts maintenance attempts for ${schemaname}.${table_name_only} failed. \
				Max_run_time will be set to $overwrite_run_time seconds\n" >> $LOG
				# Only reset timeout_seconds if it is less than the overwrite_run_time
				if (( $timeout_seconds < $overwrite_run_time )); then
					timeout_seconds=${overwrite_run_time}
				fi
			fi
			$PSQL -U $database_user $database_name <<-EOF >> $LOG 2>&1
			SET SEARCH_PATH to bp;
			UPDATE bp.bpdb_maint_driver SET max_elapsed_time = ${timeout_seconds}
			WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';
			\q
			EOF
			#
			# If we are doing maint on a catalog table, only wait 5 mins to fail
			#
			if [[ ${table_name_only} == *"pg_"* ]]; then
				timeout_seconds=300
				printf "INFO: Overwriting timeout for ${table_name_only} catalog table to ${timeout_seconds}...\n" >> $LOG
			fi
			#
			# Only cancel holding processes if we are trying to address xid consumption
			#
			if [[ "${xid_maint_fl}" == "t" ]]; then
				cancel_holding_process
			fi
			#
			# If centos6 or greater use the timeout command otherwise
			# use sleep and kill logic
			#
			if [ $cent_ver -ge 6 ]; then
				nohup /usr/bin/timeout ${timeout_seconds}s $table_file_name >> $LOG 2>&1 &	
			else
				#
				# For cent5 send the maint script to background and spawn a second monitoring script
				# this is similar to how timeout actually works
				#
				nohup $table_file_name >> $LOG 2>&1 &
				table_file_name_pid=$!
				/bin/touch ${table_file_name}_mon
				printf "#!/bin/bash \n" > ${table_file_name}_mon
				printf "run_time=0 \n" >> ${table_file_name}_mon
				printf "sleep_time=60 \n" >> ${table_file_name}_mon
				printf "pid_found=0 \n" >> ${table_file_name}_mon
				printf "pid_num=\`ps -ef|grep -w ${table_file_name}|grep -v grep|awk '{print \$2}'\` \n" >> ${table_file_name}_mon
				printf "while (( \${run_time} < ${timeout_seconds} )); do \n" >> ${table_file_name}_mon
				printf "sleep \${sleep_time} \n" >> ${table_file_name}_mon
				printf "(( run_time=runt_time+\${sleep_time} )) \n" >> ${table_file_name}_mon
				printf "pid_found=\`ps -ef|grep -w ${table_file_name}|grep -v grep|wc -l \` \n" >> ${table_file_name}_mon
				printf "if (( \${pid_found} == 0 )); then \n" >> ${table_file_name}_mon
				printf "break \n" >> ${table_file_name}_mon
				printf "fi \n" >> ${table_file_name}_mon
				printf "done \n" >> ${table_file_name}_mon
				printf "pid_found=\`ps -ef|grep -v grep|grep -cw ${table_file_name} \` \n" >> ${table_file_name}_mon
				printf "if (( \${pid_found} != 0 )); then \n" >> ${table_file_name}_mon
				printf "/bin/kill -TERM \${pid_num} >> $LOG 2>&1 \n" >> ${table_file_name}_mon
				printf "fi \n" >> ${table_file_name}_mon
				/bin/chmod 755 ${table_file_name}_mon
				nohup ${table_file_name}_mon >> $LOG 2>&1 &
			fi
		else
			$PSQL -U $database_user $database_name -tqc \
			"UPDATE bp.bpdb_maint_driver SET current_status = 'd', end_time = extract(epoch from now())::bigint \
			,comment = '-Maintenance deffered, out of time-'||comment \
			WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';" >> $LOG 2>&1
		fi
	fi
}

check_process_count (){
	#
	# Don't run more then tne number of max_maint_process_count
	# We actually get two results per process so need to divide by 2
	# to get the real count
	#
	maint_process_count=`ps -aef|grep -v grep|grep -c "${table_file_name_prefix}"`
	(( maint_process_count=maint_process_count/2 ))
	while (( $maint_process_count >= $max_maint_process_count ))
	do
		sleep 30
		check_maint_processes
		maint_process_count=`ps -aef|grep -v grep|grep -c "${table_file_name_prefix}"`
		(( maint_process_count=maint_process_count/2 ))
	done
}

cancel_holding_process (){
	printf "INFO: Attempting to cancel holding processes to run VACUUM for xid consumption\n" >> $LOG
	$PSQL -U $database_user $database_name <<-EOF >>$LOG 2>&1
		SET SEARCH_PATH to ${schemaname};
		SELECT relation::regclass,pid,client_port,backend_start,xact_start,
		query_start,state_change,waiting,state,query 
		FROM pg_locks JOIN pg_stat_activity USING (pid) 
		WHERE relation::regclass = '${table_name}'::regclass;
		SELECT pg_cancel_backend(pid)
		FROM pg_locks JOIN pg_stat_activity USING (pid) 
		WHERE relation::regclass = '${table_name}'::regclass AND pid <> pg_backend_pid() AND
		mode in ('ShareUpdateExclusiveLock','ShareLock','ShareRowExclusiveLock','ExclusiveLock','AccessExclusiveLock');
		SELECT pg_terminate_backend(pid)
		FROM pg_locks JOIN pg_stat_activity USING (pid) 
		WHERE relation::regclass = '${table_name}'::regclass AND pid <> pg_backend_pid() AND
		mode in ('ShareUpdateExclusiveLock','ShareLock','ShareRowExclusiveLock','ExclusiveLock','AccessExclusiveLock');
		\q
		EOF
	rc=$?
	check_sql_rc
}

get_pending_count (){
	#
	# Check to see if we still have tables to process in the bp.bpdb_maint_driver table
	# Check for pending or critical status
	#
	pending_maint_fl=`$PSQL -U $database_user $database_name -tqc \
	"select exists(select 1 from bp.bpdb_maint_driver where current_status in ('p','c') or current_status is null)"|head -1`
	pending_maint_fl=${pending_maint_fl# }
	# Get the row count for the driver table and use it to exit the loop just in case we have issues
	pending_maint_count=`$PSQL -U $database_user $database_name -tqc \
	"select count(1) from bp.bpdb_maint_driver"`
}
check_xid_elapsed (){
	#
	# call get_pending_count to get total row count
	#
	get_pending_count
	# If it is true that we have a pending or critical table still in the queue proceed
	while [[ "${pending_maint_fl}" == "t" ]] && (( ${pending_maint_count} > 0 )); do
		(( pending_maint_count=pending_maint_count-1 ))
		#
		# NOTE: This table name is in the format of schemaname.table_name
		# get_maint_table_name(stop_epoch bigint,bloat_threshold int4,autovac_threshold int4,rel_size_threshold bigint)
		#
		if [[ "${reason}" == "Xid" ]]; then
			table_name=`$PSQL -U $database_user $database_name -tq <<-EOF 2>>$LOG
					SET search_path to bp; 
					SELECT bp.get_maint_table_name($stop_epoch,${bloat_pct_threshold},${xid_threshold},${rel_size_threshold}); 
					\q 
					EOF`
			rc=$?
			check_sql_rc
		else
			table_name=`$PSQL -U $database_user $database_name -tq <<-EOF 2>>$LOG
					SET search_path to bp; 
					SELECT bp.get_maint_table_name($stop_epoch,${bloat_pct_threshold},${xid_threshold},${rel_size_threshold},${sample_size},${growth_rate_threshold}); 
					\q 
					EOF`
			rc=$?
			check_sql_rc
		fi
		# Clean up any empty spaces
		table_name=${table_name## }
		schemaname=${table_name%.*}
		table_name_only=${table_name#*.}
		if [[ "${table_name}" != "" ]]; then
			LOG="${out_dir}/${script_name%.*}.log"
			printf "BEGIN: `date` \n" >> $LOG 
			printf "INFO: Running special maint because of ${reason} for ${table_name} \n" >> $LOG
			printf "INFO: Maintenance should stop by $stop_epoch_date \n" >> $LOG
			if [[ "${reason}" == "Xid" ]]; then
				$PSQL -U $database_user $database_name -tqc \
				"UPDATE bp.bpdb_maint_driver SET \
				comment = '-${reason} at or above ${xid_threshold}-'||comment \
				WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';" >> $LOG 2>&1
				#
				# If we do have a table that met the xid threshold, terminate any processes
				# holding conflicting locks on the table before attempting to run the maintenance
				#
				xid_maint_fl="t"
				run_table_maint
				check_process_count
			else
				elapsed_maint_fl="t"
				$PSQL -U $database_user $database_name -tqc \
				"UPDATE bp.bpdb_maint_driver SET \
				comment = '-${reason} at or above ${growth_rate_threshold} growth rate-'||comment \
				WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';" >> $LOG 2>&1
				run_table_maint
				check_process_count
			fi
		fi
		pending_maint_fl=`$PSQL -U $database_user $database_name -tqc \
		"select exists(select 1 from bp.bpdb_maint_driver where current_status in ('p','c') or current_status is null)"|head -1`
		pending_maint_fl=${pending_maint_fl# }
		xid_maint_fl="f"
		elapsed_maint_fl="f"
	done
}

estimate_elapsed (){
	#
	# Use hdparm to get estimated throughput in MB/s
	# Just get the estimated metric once per reboot
	#
	if [[ -e /tmp/db_maint_driver_hdparm.out ]]; then
		# File is removed when bp_rcscript stop is called
		estimated_db_performance=`cat /tmp/db_maint_driver_hdparm.out`
	else
		# File is removed when bp_rcscript stop is called
		estimated_db_performance=`/sbin/hdparm -t ${db_partition}`
		#
		# Parse the hdparm output
		#
		estimated_db_performance=${estimated_db_performance##*=}
		estimated_db_performance=${estimated_db_performance%% MB*}
		estimated_db_performance=${estimated_db_performance%%.*}
		estimated_db_performance=${estimated_db_performance## }

		#
		# Convert the output from MB to BYTES
		#
		(( estimated_db_performance=estimated_db_performance*1024*1024 ))
		echo $estimated_db_performance > /tmp/db_maint_driver_hdparm.out
	fi
}

#
# Main function.
#
run_maint (){
	funcname=$FUNCNAME

	#
	# Adjust the schedule according the run_otions
	#

	# Call estimate regardless of execution options
	# It will only run hdparm if the /tmp file doesn't exist
	estimate_elapsed

	if [[ "${run_option}" =~ ^(i|m)$ ]]; then
		(( stop_epoch=stop_epoch+max_interactive_run_time ))
		stop_epoch_date=`date -d @${stop_epoch}`
	else
		(( stop_epoch=stop_epoch+scheduled_run_time ))
		stop_epoch_date=`date -d @${stop_epoch}`
		check_schedule
		#
		# If we didn't meet the schedule requirements then check for
		# xid or performance issues
		#
		if [[ "${schedule_fl}" == "f" ]]; then
			printf "BEGIN: `date` \n" >> $LOG
			printf "INFO: Running $run_verbiage \n" >> $LOG
			printf "INFO: Maintenance should stop by $stop_epoch_date \n" >> $LOG

			# populate_hst_table
			populate_driver_table

			#
			# Execute the get_maint_table_name function to get the name of a table that meets the xid
			# consumption threshold only. Set bloat threshold very high so is ignored
			# This is regardless of the schedule set in master.ini
			# This logic should reduce the need for customized schedules because of xid consumption
			#
			(( stop_epoch=stop_epoch+max_interactive_run_time ))
			stop_epoch_date=`date -d @${stop_epoch}`

			# Set bloat_pct_threshold high as to only trigger the xid threshold
			bloat_pct_threshold=120
			reason="Xid"
			check_xid_elapsed
			xid_maint_fl="f"

			#
			# Check for growth rate elapsed threshold
			# Now set xid threshold high so as to only trigger elapsed
			#

			# populate_hst_table
			populate_driver_table
			xid_threshold=120
			reason="Elapsed"
			check_xid_elapsed
			elapsed_maint_fl="f"

			#
			# Keep monitoring any maintenance jobs still running even after exiting the main loop
			# Set max_maint_process_count to 1
			#
			max_maint_process_count=1
			check_process_count
			#
			# Make sure we wait for any background processes to finish before exiting
			#
			wait

			cleanup
			printf "Exit 5\n" >> $LOG
			exit 5
		fi
	fi

	# If we passed the check_schedule logic or was passed a special flag, set the LOG file
	LOG="${out_dir}/${script_name%.*}.log"
	xid_maint_fl="f"
	elapsed_maint_fl="f"

	printf "BEGIN: `date` \n" >> $LOG
	printf "INFO: Running $run_verbiage \n" >> $LOG
	printf "INFO: Maintenance should stop by $stop_epoch_date \n" >> $LOG

	# Attempt to create the pg_repack extension
	if [ $cent_ver -ge 6 ]; then
		#
		# Check to see if pg_repack is available to be installed
		# If it is, do a quick cleanup just in case of a previous failure
		#
		pg_repack_fl=`$PSQL -U $database_user $database_name -tqc \
		"SELECT EXISTS(SELECT 1 FROM pg_available_extensions WHERE name = 'pg_repack')"|head -1`
		# Cleanup any white spaces
		pg_repack_fl=${pg_repack_fl# }
		pg_repack_fl=${pg_repack_fl% }
		if [[ "${pg_repack_fl}" == "t" ]]; then
			printf "INFO: Attempting to re-create the pg_repack extension...\n" >> $LOG
			$PSQL -U $database_user -d $database_name -c "DROP EXTENSION IF EXISTS pg_repack CASCADE;" >> $LOG 2>&1
			check_sql_rc
			$PSQL -U $database_user -d $database_name -c "CREATE EXTENSION IF NOT EXISTS pg_repack;" >> $LOG 2>&1
			check_sql_rc
		fi
	else
		printf "INFO: The pg_repack package is not installed...\n" >> $LOG
	fi

	#populate_hst_table
	populate_hst_table
	populate_driver_table

	#
	# Stop bp services if necessary
	#
	if [[ "${restart_bp_fl}" == "t" ]]; then
		stop_services
		/usr/bin/sudo -iu postgres /usr/bin/pg_ctl start -D $database_dir >> $LOG 2>&1
		check_bpdb_is_available
	fi

	#
	# Keep grabing tables that qualifiy for maintenance while we don't exceed the stop_epoch
	# call get_pending_count to get total row count and only loop thorugh table count
	#
	get_pending_count
	current_epoch=`date '+%s'`
	while (( $current_epoch < $stop_epoch )) && (( ${pending_maint_count} > 0 )); do
		(( pending_maint_count=pending_maint_count-1 ))
		#
		# Check to see if we still have talbes to process in the bp.bpdb_maint_driver table
		#
		pending_maint_fl=`$PSQL -U $database_user $database_name -tqc \
		"select exists(select 1 from bp.bpdb_maint_driver where current_status in ('p','c') or current_status is null)"|head -1`
		pending_maint_fl=${pending_maint_fl# }
		# If it is true that we have a pending or critical table still in the queue proceed
		if [[ "${pending_maint_fl}" == "t" ]]; then
			#
			# Execute the get_maint_table_name function to get the name of the next table to do maintenance on
			#
			table_name=`$PSQL -U $database_user $database_name -tq <<-EOF 2>>$LOG
					SET search_path to bp; 
					SELECT bp.get_maint_table_name($stop_epoch,${bloat_pct_threshold},${xid_threshold},${rel_size_threshold},${sample_size},${growth_rate_threshold}); 
					\q 
					EOF`
			rc=$?
			if (( $rc != 0 )); then
				printf "ERROR: Errors found running the run_maintenane_sp function...\n" >> $LOG
				printf "INFO: Will stop running the maintenance now...\n" >> $LOG
				break	
			fi
			check_sql_rc
			# Clean up any empty spaces
			table_name=${table_name## }

			#
			# NOTE: This table name is in the format of schemaname.table_name
			#
			if [[ "${table_name}" != "" ]]; then
				run_table_maint
				check_process_count
			fi
		else
			break
		fi
		current_epoch=`date '+%s'`
	done
	
	#
	# Keep monitoring any maintenance jobs still running even after exiting the main loop
	# Set max_maint_process_count to 1
	#
	max_maint_process_count=1
	check_process_count

	#
	# If maint option then just update all rows current_status to m
	#
	if [[ "${run_option}" == "m" ]]; then
		$PSQL -U $database_user $database_name -tqc"UPDATE bp.bpdb_maint_driver SET current_status = 'm', end_time = extract(epoch from now())::bigint;" >> $LOG  2>&1
	fi

	#
	# Make sure to wait for any background shell processes possibly still running
	#
	wait

	#
	# Archive all current rows in the bpdb_maint_driver table
	#
	populate_hst_table

	#
	# Restart bp services if necessary
	#
	if [[ "${restart_bp_fl}" == "t" ]]; then
		start_services
	fi
	
	#
	# If the run_option is i then show stats after run
	#
	if [[ "${run_option}" =~ ^(i)$ ]]; then
		if [[ "${table_name_list}" != "" ]]; then
			populate_driver_table
			for table_name in `echo $table_name_list`
			do
				schemaname=${table_name%.*}
				table_name_only=${table_name#*.}
				printf "Stats for $table_name table maintenance.\n"|tee -a $LOG
				$PSQL -U $database_user $database_name <<-EOF |tee -a $LOG
					\qecho After maintenance:
					SELECT relname,reltuples as num_rows,
					pg_size_pretty(total_rel_size) as rel_size,
					relpages,expectedpages,
					bloatpct,autovac_freeze_pct
					FROM bp.bpdb_maint_driver
					WHERE schemaname = '${schemaname}' AND relname = '${table_name_only}';
					EOF
			done
			#
			# Cleanup bpdb_maint_driver table and save the latest stats using current_status = m for maintenance
			#
			$PSQL -U $database_user $database_name -tqc"UPDATE bp.bpdb_maint_driver SET current_status = 'm', end_time = extract(epoch from now())::bigint;" >> $LOG  2>&1
			populate_hst_table
		else
			printf "No tables qualified at this time for maintenance.\n" |tee -a $LOG
		fi
	fi
}

#
#main script
#

# 
# Even if the -l option is not used log acivity if schedule qualifies
# strip the .sh from the script name for the log name
#
LOG="${out_dir}/${script_name%.*}.log"
/bin/touch $LOG

#
# Check to see if we have yum processes running
#
found_yum=`/bin/ps -ef |grep -v grep|grep -cE 'rpm|yum'`
if [ $found_yum -ne 0 ]; then
	printf "WARNING: Found yum or rpm processes running...\n" >> $LOG
	printf "Exiting now...\n" >> $LOG
	exit 8
fi


#
# Check to see if we have maintenance processes running
# Check for individual maint script or the main db_maint script
#
/usr/bin/pgrep ${table_file_name_prefix} > /dev/null 2>&1
found_script=$?
if [ $found_script -eq 0 ]; then
	printf "WARNING: Found maintenance processes running...\n" >> $LOG
	printf "Exiting now...\n" >> $LOG
	exit 8
fi

#
# Exclude processes used to monitor the log of the script
#
found_script=`/bin/ps -ef|grep ${script_name%.*}|grep -c -v "${maint_pid}\|grep\|vi\|tail\|cat"`
if [ $found_script -ne 0 ]; then
	printf "WARNING: Found maintenance processes running...\n" >> $LOG
	printf "Exiting now...\n" >> $LOG
	exit 8
fi

#
# If the lock file is older than max_interactive_run_time then try to remove it
# Maybe a catastrophic event happend
#
if [[ -e ${lockFile} ]]; then
	if test "`find ${lockFile} -mmin +${max_interactive_run_time}`"; then
		printf "Found ${lockFile}.  Will remove it now...\n" >> -a $LOG
		/bin/rm -f ${lockFile} >/dev/null 2>&1
	fi
fi

#
#  Make sure only one copy of this script can run at a time.
#  Only lock the file no more than max_interactive_run_time
#
if ! lockfile -r0 -l $max_interactive_run_time "$lockFile"; then
	printf "${script_name}: Can't get the lock\n" >> $LOG
	exit 8
fi

#
# Make sure we can find the psql command or exit
#
check_psql
rc=0

#
# Check is bpdb is available
#
check_bpdb_is_available


#
# Reset log just in case -l option was not used
#
LOG="/dev/null"

#
# Make sure we don't have old /tmp files laying around
#
/bin/rm -f /tmp/${table_file_name_prefix}* >/dev/null 2>&1


#
# Check if DBMaintForce is set to true
#
if [[ $force_db_maint == "t" ]]; then
	restart_bp_fl="t"
	run_verbiage="FORCE ${run_verbiage}"
fi

#
# Use the DBMaintMem percentage and only overwrite maintenance_work_mem if necessary
# In other words, don't lower maintenance_work_mem, only increase if it is necessary
# increase it
#
MEM_KB=`grep MemTotal /proc/meminfo | grep -E -o "[0-9]+"`
# get the percentage of memory in KB we will use for maintenance_work_mem
maintenance_work_mem=`echo $MEM_KB $maintenance_work_mem | awk '{printf "%lu", $1*$2}'`

default_maintenance_work_mem=`${PSQL} -U ${database_user} -tc"SELECT setting::BIGINT FROM pg_settings WHERE name = 'maintenance_work_mem'"`
default_maintenance_work_mem=${default_maintenance_work_mem# }
default_maintenance_work_mem=${default_maintenance_work_mem% }
# if the maintenance_work_mem in the config file is less overwrite it
if (( ${default_maintenance_work_mem} < ${maintenance_work_mem} )); then
	maintenance_work_mem=$default_maintenance_work_mem
fi

#
# Estimated database partition performance
#
db_repo=`${PSQL} -U ${database_user} -d ${database_name} -c"SHOW data_directory;" | sed -n 3p`
db_partition=`df -PhT $db_repo|tail -1|awk '{print $1}'`
db_repo_total_kb=$(df -Pk $db_repo | tail -1 | awk '{print $2}')
db_repo_used_kb=$(df -Pk $db_repo | tail -1 | awk '{print $3}')
db_repo_free_kb=$(df -Pk $db_repo | tail -1 | awk '{print $4}')

if (( $# < 1 )); then
        run_maint
	cleanup
        exit $rc
fi

if [[ $1 == *"im"* || $1 == *"mi"* || $1 == *"fm"* || $1 == *"mf"* ]]; then
	printf "Invalid option passed.\n $USAGE"
	cleanup
	exit 1
fi

while getopts lmif opt
do
        case "$opt" in
                i) run_option="i${run_option}";
                   run_verbiage="INTERACTIVE ${run_verbiage}";;
                m) run_option="m${run_option}";
                   run_verbiage="MONITORING ${run_verbiage}";;
                l) run_option="l${run_option}";
		   LOG="${out_dir}/${script_name%.*}.log";
		   run_verbiage="LOGGING ${run_verbiage}";;
                f) restart_bp_fl="t";
		   run_verbiage="FORCE ${run_verbiage}";;
                *) printf "Invalid option passed.\n $USAGE";
		   cleanup;
                   exit 1;;
        esac
done

if [[ "${run_option}" == *"m"* ]]; then
        run_option="m"
else
	if [[ "${run_option}" == *"i"* || $restart_bp_fl == "t" ]]; then
		run_option="i"
	fi
fi

case "$run_option" in
        l) run_maint; cleanup; exit $rc;;
        m) run_maint; cleanup; exit $rc;;
        i) run_maint; cleanup; exit $rc;;
	*) printf "Invalid option passed.\n $USAGE";
	   cleanup;
	   exit 1;;
esac
