#!/bin/bash
UL="\033[4m"
EUL="\033[0m"

USAGE="
Usage:
	$(basename $0) user [${UL}username${EUL}...]
	$(basename $0) pass ${UL}username${EUL}...
	$(basename $0) host [${UL}hostname${EUL}...]
"	

HELP="
Configures security options for Exchange FLR.

user	Sets user(s) that may access the Exchange FLR share. No ${UL}username${EUL}
	will print the current setting. A ${UL}username${EUL} of 'guest' will
	allow unauthenticated access to the share. This will create any
	${UL}username${EUL}	that is unknown to the system.

pass	Sets the samba password for each ${UL}username${EUL}. This will create
	any ${UL}username${EUL} that is unknown to the system.

host	Sets host(s) that may access the Exchange FLR share. No ${UL}hostname${EUL}
	will print the current setting. ${UL}hostname${EUL} of 'guest' will allow
	unrestricted access to the share.
"

exchange_args_check () {
	for arg in $*
	do
		if [ -z "${arg##*\#*}" -o -z "${arg##*\;*}" ]
		then
			echo "Invalid argument ${arg}"
			exit 1
		fi
	done
}

exchange_user_security () {
	sed -i 's/^[ \t]*security = share[ \t]*$/\tsecurity = user/' /etc/samba/smb.conf
	if grep -q "map to guest" /etc/samba/smb.conf
	then
		sed -i 's/^[ \t]*map to guest = [a-z ]*[ \t]*$/\tmap to guest = bad user/' /etc/samba/smb.conf
	else
		cp /etc/samba/smb.conf /etc/samba/.smb.conf.bak
		awk '{print}; /security =/ { print "\tmap to guest = bad user"}' /etc/samba/.smb.conf.bak > /etc/samba/smb.conf
	fi
}

exchange_user_password () {
	if [ $# != 1 ]
	then
		echo "Invalid number of args, expected 1: ($#) $*"
		return
	fi

	echo "Setting password ${user}"
	smbpasswd -a $1
}

exchange_user_create () {
	for user in $*
	do
		if [ "${user}" == "guest" ]
		then
			continue
		fi

		if ! id ${user} > /dev/null 2>&1
		then
			echo "Creating user ${user}"
			if useradd -M -d / -s /sbin/nologin ${user}
			then
				exchange_user_password ${user}
			else
				echo "Could not create user: ${user}"
				exit 1
			fi
		fi
	done
}

exchange_user_get () {
	bputil -g "ExchangeFLR" "ValidUsers" "guest" /usr/bp/bpinit/master.ini
}

exchange_user_set () {
	bputil -p "ExchangeFLR" "ValidUsers" "$*" /usr/bp/bpinit/master.ini
}

exchange_host_get () {
	bputil -g "ExchangeFLR" "HostsAllow" "guest" /usr/bp/bpinit/master.ini
}

exchange_host_set () {
	bputil -p "ExchangeFLR" "HostsAllow" "$*" /usr/bp/bpinit/master.ini
}

exchange_user () {
	if [ $# -eq 0 ]
	then
		exchange_user_get
	else
		exchange_user_security
		exchange_user_create $*
		exchange_user_set $*
	fi
}

exchange_pass () {
	for user in $*
	do
		if ! id ${user} > /dev/null 2>&1
		then
			exchange_user_create ${user}
		else
			exchange_user_password ${user}
		fi
	done
}

exchange_host () {
	if [ $# -eq 0 ]
	then
		exchange_host_get
	else
		exchange_host_set $*
	fi
}

ACTION=$1
shift 1

exchange_args_check $*

case ${ACTION} in
	user)
		exchange_user $*
		;;
	pass)
		exchange_pass $*
		;;
	host)
		exchange_host $*
		;;
	help)
		echo -e "${USAGE}\n${HELP}"
		exit
		;;
	*)
		echo -e "${USAGE}\n${HELP}"
		exit 1
		;;
esac



