<?php

class PoliciesReport
{
    private $BP;
    const UNKNOWN_ELEMENT = "Unknown";

    // Linux volume preface
    const LINUX_VOLUME = '@@@:';

    public function __construct($BP, $functionsClass)
    {
        $this->BP = $BP;
        $this->functions = $functionsClass;
        $this->localID = $this->BP->get_local_system_id();
        $this->showSLAPolicies = $this->functions->showSLAPolicies($this->localID);
    }

    /*
     * Main routine that drives the protection policies report.
     */
    public function get_jobs_for_all_assets($systems)
    {

        $policiesReport = array();
        $reportRows = array();

        // Get a list of all jobs on the system.
        $coldBackupCopyList = $this->getAllSchedules('archive', $systems);
        $hotBackupCopyList = $this->getAllSchedules('replication', $systems);
        $scheduleList = $this->getAllSchedules('backup', $systems);
        $policyAssets = $this->getPolicyAssets($systems);

        // The backup job list is the driver of the report.
        foreach ($scheduleList as $schedule) {
            if (!isset($schedule['enabled']) || !$schedule['enabled']) {
                continue;
            }
            $sid = $schedule['sid'];
            if (isset($schedule['instances'])) {
                $instances = $schedule['instances'];
                foreach ($instances as $instance) {
                    $instance['retention'] = $this->getInstanceRetention($instance, $schedule['sid']);
                    $reportRow = $instance;
                    $reportRow['app_name'] = $schedule['application_name'];
                    $reportRow['backup_job_name'] = $schedule['name'];
                    $reportRow['archive_job_name'] = $this->getJobsForInstance($instance['id'], $coldBackupCopyList, $sid);
                    $reportRow['replication_job_name'] = $this->getJobsForInstance($instance['id'], $hotBackupCopyList, $sid);
                    if ($sid === $this->localID) {
                        // Names only available if local.
                        $reportRow['policy_name'] = $this->getPolicyForInstance($instance['id'], $policyAssets);
                    } else {
                        $reportRow['policy_name'] = 'Visible on local appliance only';
                    }
                    $reportRow['types'] = $this->getTypes($schedule);
                    $reportRow['system_name'] = $this->getSystemName($systems, $sid);
                    $reportRow['details'] = $this->getDetails($instance, $schedule);
                    unset($reportRow['job_info']['instances']);
                    unset($reportRow['retention']);
                    $reportRows[] = $reportRow;
                }
            } else if (isset($schedule['clients'])) {
                $this->sanitizeInclExclLists($schedule['clients']);
                $clients = $schedule['clients'];
                foreach ($clients as $client) {
                    $client['retention'] = $this->getClientRetention($client, $sid);
                    $client['client_name'] = $client['name'];
                    $clientInfo = $this->BP->get_client_info($client['id'], $sid);
                    $client_instance = ($clientInfo !== false) ? $clientInfo['file_level_instance_id'] : null;
                    $client['client_id'] = $client['id'];
                    $client['id'] = $client_instance;
                    $reportRow = $client;
                    $reportRow['app_name'] = $schedule['application_name'];
                    $reportRow['backup_job_name'] = $schedule['name'];
                    $reportRow['archive_job_name'] = $this->getJobsForInstance($client_instance, $coldBackupCopyList, $sid);
                    $reportRow['replication_job_name'] = $this->getJobsForInstance($client_instance, $hotBackupCopyList, $sid);
                    if ($sid === $this->localID) {
                        // Names only available if local.
                        $reportRow['policy_name'] = $this->getPolicyForInstance($client_instance, $policyAssets);
                    } else {
                        $reportRow['policy_name'] = 'Visible on local appliance only';
                    }
                    $reportRow['types'] = $this->getTypes($schedule);
                    $reportRow['system_name'] = $this->getSystemName($systems, $sid);
                    $reportRow['details'] = $this->getDetails($client, $schedule);
                    unset($reportRow['job_info']['clients']);
                    unset($reportRow['retention']);
                    $reportRows[] = $reportRow;
                }
            }
        }

        $policiesReport['count'] = count($reportRows);
        $policiesReport['total'] = count($reportRows);
        $policiesReport['data'] = $reportRows;

        return $policiesReport;
    }

    /*
     * Given a systems list and the ID, get the name.
     */
    private function getSystemName($systems, $sid) {
        return $systems[$sid];
    }

    /*
     * Get all the schedules at the top level (without "info"), equivalent to GET /api/joborders.
     * For each job, get the "info", equivalend of GET /api/joborders/{id} and merge the results.
     */
    private function getAllSchedules($type, $systems) {
        global $schedules;
        $list = array();
        $scheduleList = $schedules->get($type, $systems);
        if ($scheduleList !== false) {
            $scheduleList = $scheduleList['data'];
            foreach ($scheduleList as $schedule) {
                $detailedData = $schedules->get($schedule['id'], $systems);
                if ($detailedData !== false && is_array($detailedData) && count($detailedData) > 0) {
                    $detailedScheduleInfo = $detailedData['data'][0];
                    $schedule = array_merge($schedule, $detailedScheduleInfo);
                }
                $list[] = $schedule;
            }
        }
        return $list;
    }

    /*
     * Get the SLA protected assets on the system, returns none (empty array) if not showing policies.
     */
    private function getPolicyAssets($systems) {
        $policyAssets = array();
        if ($this->showSLAPolicies) {
            global $SLAPolicy;
            $policyAssets = $SLAPolicy->get('assets', false, $systems);
            $policyAssets = $policyAssets['data'];
        }
        return $policyAssets;

    }

    /*
     * Given a list of jobs and an instance ID, see if this instance is in the job.
     */
    private function getJobsForInstance($instance_id, $jobList, $sid) {
        $names = "";
        if (is_array($jobList)) {
            //global $Log;
            //$Log->writeVariable("Searching for...... " . $instance_id . " for job with sid " . $sid);
            //$Log->writeVariable($jobList);
            $nameArray = array();
            foreach ($jobList as $job) {
                if ($job['sid'] == $sid) {
                    // Is the job on the same system as the instance?
                    $instances = $job['instances'];
                    foreach ($instances as $instance) {
                        if ($instance_id == $instance['id']) {
                            //$Log->writeVariable("FOUND, instance " . $instance_id . " in job " . $job['name']);
                            $nameArray[] = $job['name'];
                        }
                    }
                }
            }
            $names = implode(',', $nameArray);
        }
        return $names;
    }

    /*
     * If the instance is protected by a policy, get its name.
     */
    private function getPolicyForInstance($instance_id, $policyAssets) {
        $policyName = null;
        if ($this->showSLAPolicies) {
            foreach ($policyAssets as $asset) {
                if ($asset['id'] == $instance_id) {
                    $policyName = $asset['policy_name'];
                }
            }
        } else {
            $policyName = "Not Applicable";
        }
        return $policyName;
    }

    /*
     * Get the backup types included in this calendar.
     */
    private function getTypes($schedule) {
        $types = "";
        if (isset($schedule['calendar'])) {
            $calendar = $schedule['calendar'];
            $typeArray = array();
            if (is_array($calendar) && count($calendar) > 0) {
                foreach ($calendar as $item) {
                    if (isset($item['backup_type'])) {
                        $typeArray[] = $item['backup_type'];
                    }
                }
                $types = implode(',', $typeArray);
            }
        }
        return $types;
    }

    /*
     * Gets the details that will be seen when users click on a report row.
     */
    private function getDetails($object, $schedule) {

        $details = array();

        $details['backup_job_description'] = $schedule['description'];
        $details['includes'] = isset($object['incl_list']) ? $object['incl_list'] : array();
        $details['excludes'] = isset($object['excl_list']) ? $object['excl_list'] : array();
        $details['job_name'] = $schedule['name'];
        $details['calendar_str'] = $schedule['calendar_str'];
        $details['types'] = $this->getTypes($schedule);
        $details['name'] = $object['name'];
        $details['client_name'] = $object['client_name'];
        $details['id'] = $object['id'];
        $details['client_id'] = isset($object['client_id']) ? $object['client_id'] :
                                (isset($schedule['client_id']) ? $schedule['client_id'] : null);
        $details['retention_min'] = $object['retention']['retention_min'];
        $details['retention_max'] = $object['retention']['retention_max'];
        $details['legal_hold'] = $object['retention']['legal_hold'];

        return $details;
    }

    /*
     * Get the retention for the specified instance.
     */
    private function getInstanceRetention($instance, $sid) {

        $retention = array('retention_min' => 'N/A', 'retention_max' => 'N/A', 'legal_hold' => 'N/A');

        $filter = array('instance_id' => $instance['id']);
        $retentionInfo = $this->BP->get_retention_settings($filter, $sid);
        if ($retentionInfo !== false) {
            if (count($retentionInfo) > 0) {
                $retention = $retentionInfo[0];
                $instanceRetention = array('retention_min' => $retention['retention_min'],
                    'retention_max' => $retention['retention_max'],
                    'legal_hold' => $retention['legal_hold']);
                $retention = $instanceRetention;
            }
        } else {
            // set retention error, will show as "N/A".
            global $Log;
            $Log->writeError("Error getting retention info", true);
        }
        return $retention;
    }

    /*
     * Get the retention for the specified client.
     */
    private function getClientRetention($client, $sid) {

        $retention = array('retention_min' => 'N/A', 'retention_max' => 'N/A', 'legal_hold' => 'N/A');

        // filter to the client ID and the file-level app ID.
        $filter = array('client_id' => $client['id'], 'app_id' => Constants::APPLICATION_ID_FILE_LEVEL);
        $retentionInfo = $this->BP->get_retention_settings($filter, $sid);
        if ($retentionInfo !== false) {
            if (count($retentionInfo) > 0) {
                $retention = $retentionInfo[0];
                $instanceRetention = array('retention_min' => $retention['retention_min'],
                    'retention_max' => $retention['retention_max'],
                    'legal_hold' => $retention['legal_hold']);
                $retention = $instanceRetention;
            }
        } else  {
            // set retention error, will show as "N/A".
            global $Log;
            $Log->writeError("Error getting retention info", true);
        }
        return $retention;
    }

    /*
     * Remove any '@@@' preface from inclusion/exclusion lists.
     */
    function sanitizeInclExclLists(&$clients) {
        foreach ($clients as &$client) {
            if (isset($client['incl_list'])) {
                foreach ($client['incl_list'] as &$fileName) {
                    $fileName = str_replace(self::LINUX_VOLUME, "", $fileName);
                }
            }
            if (isset($client['excl_list'])) {
                foreach ($client['excl_list'] as &$fileName) {
                    $fileName = str_replace(self::LINUX_VOLUME, "", $fileName);
                }
            }
        }
    }
}
?>
